/**
 *
 */


// System Import Headers.
#include <ctype.h>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <iterator>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>

// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "structs.h"
#endif


// Method Declaration.
path_validator_struct* validate_path(std::string path_str);
void process_dir(std::string path_str);
void process_file(std::string path_str);
// void validate_path(char* path_str);
// void compute(int a, int b);
// __global__ void cuda_add(int a, int b, int* c);


// Global Variables.


/*
 * Program's main. Initializes and runs program.
 */
int main(int argc, char* argv[]) {
    printf("Starting program.\n");
    printf("\n");

    // Check provided number of args.
    if (argc < 2) {
        // No args provided.
        printf("Provided too few program args. Expected one arg.");
    } else if (argc > 2) {
        // More than one arg provided.
        printf("Provided too many program args. Expected one arg.");
    } else {
        // Correct number of args provided. Attempt to process.

        // Check data type.
        std::string path_str = argv[1];
        path_validator_struct* return_struct = validate_path(path_str);
        int path_type = return_struct->file_type;
        if (return_struct->err_code) {
            std::cerr << "Error in is_regular_file: " << return_struct->err_code.message();
        }
        free_path_validator_struct(return_struct);

        // Handle based on provided type.
        if (path_type == TYPE_DIR) {
            // Process dir.
            process_dir(path_str);
        } else if (path_type == TYPE_FILE) {
            // Process file.
            process_file(path_str);
        }
    }

    printf("\n");
    printf("Terminating program.\n");
    exit(0);
}



/**
 * Validates provided path and if valid, determines path type.
 * Returns struct of data.
 */
path_validator_struct* validate_path(std::string path_str) {
    int type_value = TYPE_OTHER;
    std::error_code ec;

    std::cout << "path_str: \"" + path_str + "\"\n";

    // Check if provided path exists.
    std::ifstream test_path_exists(path_str);
    if (test_path_exists) {
        // Path exists. Check if dir or file.
        if (std::filesystem::is_directory(path_str, ec)) {
            // Is dir.
            type_value = TYPE_DIR;
        } else if ((! ec) && (std::filesystem::is_regular_file(path_str, ec))) {
            // Is file.
            type_value = TYPE_FILE;
        }
    }

    // Return values.
    path_validator_struct* return_struct_ptr = initialize_path_validator_struct(type_value, ec);
    return return_struct_ptr;
}


/**
 * Finds and processes all files in directory.
 * Assumes directory is full of image files.
 */
void process_dir(std::string path_str) {

    // Process all files in directory.
    for (const auto & entry : std::filesystem::directory_iterator(path_str)) {
        // Get as string.
        std::string file_path_str{entry.path().u8string()};
        // std::cout << file_path_str << std::endl;

        // Send individual file path for processing.
        process_file(entry.path());
    }
}


/**
 * Processes a single image file.
 */
void process_file(std::string path_str) {

    // Open input file.
    std::ifstream input_file;
    input_file.open(path_str, std::ios::binary | std::ios::in);
    if (! input_file.good()) {
        printf("Failed to open file \"%s\".\n", path_str);
    }

    // Calculate output path location.
    std::string base_filename = path_str.substr(path_str.find_last_of("/\\") + 1);
    std::string output_path_str = "../output/" + base_filename;

    // Open output file.
    std::ofstream output_file;
    output_file.open(output_path_str, std::ios::binary | std::ios::out);
    if (! output_file.good()) {
        printf("Failed to open file \"%s\".\n", output_path_str);
    }

    // Copy file over.
    if (input_file.good() && output_file.good()) {
        // output_file << input_file;

        // int buffer[100];
        // while (input_file.read( (char*) &buffer, sizeof(buffer)) ) {
        //     output_file.write( (char*) &buffer, sizeof(buffer) );
        // }

        std::copy(
            std::istreambuf_iterator<char>(input_file),
            std::istreambuf_iterator<char>( ),
            std::ostreambuf_iterator<char>(output_file)
        );

        input_file.close();
        output_file.close();
    }
}





// /*
//  * Program's main. Initializes and runs program.
//  */
// int main(int argc, char* argv[]) {
//     printf("Starting program.\n");
//     printf("\n");

//     // Check provided number of args.
//     if (argc < 3) {
//         // Less than two args provided.
//         printf("Provided too few program args. Expected two integers to add.");
//     } else if (argc > 3) {
//         // More than two args provided.
//         printf("Provided too many program args. Expected two integers to add.");
//     } else {
//         // Two args provided. Compute with GPU.
//         int a = atoi(argv[1]);
//         int b = atoi(argv[2]);
//         printf("Adding %i and %i\n", a, b);
//         compute(a, b);
//     }

//     printf("\n");
//     printf("Terminating program.\n");
//     exit(0);
// }


// /*
//  * Function to have CPU and GPU collaborate for basic addition computation.
//  */
// void compute(int a, int b) {
//     int sum;
//     int* dev_sum;

//     // Allocate memory on the GPU.
//     cudaMalloc((int**) &dev_sum, sizeof(int));

//     // Call CUDA function.
//     cuda_add<<<1,1>>>(a, b, dev_sum);

//     // Copy GPU result back to CPU.
//     cudaMemcpy(&sum, dev_sum, sizeof(int), cudaMemcpyDeviceToHost);

//     // Display result.
//     printf("Sum is %i\n", sum);

//     cudaFree(dev_sum);
// }


// /**
//  * Entrypoint to GPU kernel execution.
//  */
// __global__ void cuda_add(int a, int b, int* c) {
//     *c = a + b;
// }
